@@ -1,17 +1,31 @@
-local a = vim.api
-local query = require('vim.treesitter.query')
+local api = vim.api
+local query = vim.treesitter.query
+
+---@alias TSHlIter fun(): integer, TSNode, TSMetadata
+
+---@class TSHighlightState
+---@field next_row integer
+---@field iter TSHlIter|nil
 
--- support reload for quick experimentation
 ---@class TSHighlighter
+---@field active table<integer,TSHighlighter>
+---@field bufnr integer
+---@field orig_spelloptions string
+---@field _highlight_states table<TSTree,TSHighlightState>
+---@field _queries table<string,TSHighlighterQuery>
+---@field tree LanguageTree
 local TSHighlighter = rawget(vim.treesitter, 'TSHighlighter') or {}
 TSHighlighter.__index = TSHighlighter
 
 TSHighlighter.active = TSHighlighter.active or {}
 
+---@class TSHighlighterQuery
+---@field _query Query|nil
+---@field hl_cache table<integer,integer>
 local TSHighlighterQuery = {}
 TSHighlighterQuery.__index = TSHighlighterQuery
 
-local ns = a.nvim_create_namespace('treesitter/highlighter')
+local ns = api.nvim_create_namespace('treesitter/highlighter')
 
 ---@private
 function TSHighlighterQuery.new(lang, query_string)
@@ -22,7 +36,7 @@ function TSHighlighterQuery.new(lang, query_string)
       local name = self._query.captures[capture]
       local id = 0
       if not vim.startswith(name, '_') then
-        id = a.nvim_get_hl_id_by_name('@' .. name .. '.' .. lang)
+        id = api.nvim_get_hl_id_by_name('@' .. name .. '.' .. lang)
       end
 
       rawset(table, capture, id)
@@ -31,22 +45,24 @@ function TSHighlighterQuery.new(lang, query_string)
   })
 
   if query_string then
-    self._query = query.parse_query(lang, query_string)
+    self._query = query.parse(lang, query_string)
   else
-    self._query = query.get_query(lang, 'highlights')
+    self._query = query.get(lang, 'highlights')
   end
 
   return self
 end
 
----@private
+---@package
 function TSHighlighterQuery:query()
   return self._query
 end
 
---- Creates a new highlighter using @param tree
+---@package
 ---
----@param tree LanguageTree |LanguageTree| parser object to use for highlighting
+--- Creates a highlighter for `tree`.
+---
+---@param tree LanguageTree parser object to use for highlighting
 ---@param opts (table|nil) Configuration of the highlighter:
 ---           - queries table overwrite queries used by the highlighter
 ---@return TSHighlighter Created highlighter object
@@ -57,7 +73,7 @@ function TSHighlighter.new(tree, opts)
     error('TSHighlighter can not be used with a string parser source.')
   end
 
-  opts = opts or {}
+  opts = opts or {} ---@type { queries: table<string,string> }
   self.tree = tree
   tree:register_cbs({
     on_changedtree = function(...)
@@ -66,18 +82,20 @@ function TSHighlighter.new(tree, opts)
     on_bytes = function(...)
       self:on_bytes(...)
     end,
-    on_detach = function(...)
-      self:on_detach(...)
+    on_detach = function()
+      self:on_detach()
     end,
   })
 
-  self.bufnr = tree:source()
+  self.bufnr = tree:source() --[[@as integer]]
   self.edit_count = 0
   self.redraw_count = 0
   self.line_count = {}
   -- A map of highlight states.
   -- This state is kept during rendering across each line update.
   self._highlight_states = {}
+
+  ---@type table<string,TSHighlighterQuery>
   self._queries = {}
 
   -- Queries for a specific language can be overridden by a custom
@@ -88,6 +106,8 @@ function TSHighlighter.new(tree, opts)
     end
   end
 
+  self.orig_spelloptions = vim.bo[self.bufnr].spelloptions
+
   vim.bo[self.bufnr].syntax = ''
   vim.b[self.bufnr].ts_highlight = true
 
@@ -101,7 +121,7 @@ function TSHighlighter.new(tree, opts)
     vim.cmd.runtime({ 'syntax/synload.vim', bang = true })
   end
 
-  a.nvim_buf_call(self.bufnr, function()
+  api.nvim_buf_call(self.bufnr, function()
     vim.opt_local.spelloptions:append('noplainbuffer')
   end)
 
@@ -120,12 +140,14 @@ function TSHighlighter:destroy()
     vim.bo[self.bufnr].spelloptions = self.orig_spelloptions
     vim.b[self.bufnr].ts_highlight = nil
     if vim.g.syntax_on == 1 then
-      a.nvim_exec_autocmds('FileType', { group = 'syntaxset', buffer = self.bufnr })
+      api.nvim_exec_autocmds('FileType', { group = 'syntaxset', buffer = self.bufnr })
     end
   end
 end
 
----@private
+---@package
+---@param tstree TSTree
+---@return TSHighlightState
 function TSHighlighter:get_highlight_state(tstree)
   if not self._highlight_states[tstree] then
     self._highlight_states[tstree] = {
@@ -142,28 +164,31 @@ function TSHighlighter:reset_highlight_state()
   self._highlight_states = {}
 end
 
----@private
+---@package
+---@param start_row integer
+---@param new_end integer
 function TSHighlighter:on_bytes(_, _, start_row, _, _, _, _, _, new_end)
-  a.nvim__buf_redraw_range(self.bufnr, start_row, start_row + new_end + 1)
+  api.nvim__buf_redraw_range(self.bufnr, start_row, start_row + new_end + 1)
 end
 
----@private
+---@package
 function TSHighlighter:on_detach()
   self:destroy()
 end
 
----@private
+---@package
+---@param changes integer[][]?
 function TSHighlighter:on_changedtree(changes)
   for _, ch in ipairs(changes or {}) do
-    a.nvim__buf_redraw_range(self.bufnr, ch[1], ch[3] + 1)
+    api.nvim__buf_redraw_range(self.bufnr, ch[1], ch[3] + 1)
   end
 end
 
 --- Gets the query used for @param lang
 --
----@private
+---@package
 ---@param lang string Language used by the highlighter.
----@return Query
+---@return TSHighlighterQuery
 function TSHighlighter:get_query(lang)
   if not self._queries[lang] then
     self._queries[lang] = TSHighlighterQuery.new(lang)
@@ -173,6 +198,10 @@ function TSHighlighter:get_query(lang)
 end
 
 ---@private
+---@param self TSHighlighter
+---@param buf integer
+---@param line integer
+---@param is_spell_nav boolean
 local function on_line_impl(self, buf, line, is_spell_nav)
   self.tree:for_each_tree(function(tstree, tree)
     if not tstree then
@@ -207,11 +236,12 @@ local function on_line_impl(self, buf, line, is_spell_nav)
         break
       end
 
-      local start_row, start_col, end_row, end_col = node:range()
+      local range = vim.treesitter.get_range(node, buf, metadata[capture])
+      local start_row, start_col, _, end_row, end_col, _ = unpack(range)
       local hl = highlighter_query.hl_cache[capture]
 
       local capture_name = highlighter_query:query().captures[capture]
-      local spell = nil
+      local spell = nil ---@type boolean?
       if capture_name == 'spell' then
         spell = true
       elseif capture_name == 'nospell' then
@@ -222,7 +252,7 @@ local function on_line_impl(self, buf, line, is_spell_nav)
       local spell_pri_offset = capture_name == 'nospell' and 1 or 0
 
       if hl and end_row >= line and (not is_spell_nav or spell ~= nil) then
-        a.nvim_buf_set_extmark(buf, ns, start_row, start_col, {
+        api.nvim_buf_set_extmark(buf, ns, start_row, start_col, {
           end_line = end_row,
           end_col = end_col,
           hl_group = hl,
@@ -236,10 +266,13 @@ local function on_line_impl(self, buf, line, is_spell_nav)
         state.next_row = start_row
       end
     end
-  end, true)
+  end)
 end
 
 ---@private
+---@param _win integer
+---@param buf integer
+---@param line integer
 function TSHighlighter._on_line(_, _win, buf, line, _)
   local self = TSHighlighter.active[buf]
   if not self then
@@ -250,6 +283,9 @@ function TSHighlighter._on_line(_, _win, buf, line, _)
 end
 
 ---@private
+---@param buf integer
+---@param srow integer
+---@param erow integer
 function TSHighlighter._on_spell_nav(_, _, buf, srow, _, erow, _)
   local self = TSHighlighter.active[buf]
   if not self then
@@ -264,6 +300,7 @@ function TSHighlighter._on_spell_nav(_, _, buf, srow, _, erow, _)
 end
 
 ---@private
+---@param buf integer
 function TSHighlighter._on_buf(_, buf)
   local self = TSHighlighter.active[buf]
   if self then
@@ -272,6 +309,9 @@ function TSHighlighter._on_buf(_, buf)
 end
 
 ---@private
+---@param _win integer
+---@param buf integer
+---@param _topline integer
 function TSHighlighter._on_win(_, _win, buf, _topline)
   local self = TSHighlighter.active[buf]
   if not self then
@@ -283,7 +323,7 @@ function TSHighlighter._on_win(_, _win, buf, _topline)
   return true
 end
 
-a.nvim_set_decoration_provider(ns, {
+api.nvim_set_decoration_provider(ns, {
   on_buf = TSHighlighter._on_buf,
   on_win = TSHighlighter._on_win,
   on_line = TSHighlighter._on_line,
